<?php


// Turn off all error reporting
error_reporting(0);

//the request requires a $s_requestStaticSecret to be set equal to the configures $s_requestStaticSecret in the config

/*
 * 2016 - 2017 MIT Janis Dähne
 */

/*
 * version 0.9
 */

#requires sql helper
include_once './sqlHelper.php';

#requires some helper function to e.g. create / delete dirs, handle output...
require './helpers.php';

#require some constants
include_once './constants.php';

include_once './transactionHelper.php';

/**
 * checks if a file name is valid the most basic case is letters, numbers, underscore
 * @param $file a file name
 * @return bool
 */
function is_valid_fileName($file)
{
    return preg_match('/^[a-zA-Z0-9_]+$/', $file) > 0;
}

if ($isDebug) {
    $time_pre = microtime(true);
}

# checks ip and reads the config
$config = require './bootstrap.php';

if ($isDebug) {
    $time_post = microtime(true);
    $exec_time = ($time_post - $time_pre) * 1000; //in ms
    debug("time to read & decode config & check ip in ms: " . $exec_time);
}


if ($isDebug) {
    $time_pre = microtime(true);
}
#request was allowed and config was read

# ============================================================== now check if we have all arguments that we need...

$entityBody = file_get_contents('php://input');

$body = json_decode($entityBody, TRUE);

if ($body === NULL) {
    output($status_code_ArgumentInvalidOrMissing, "could not decode request body");
    exit(1);
}


if (!isset($body[$s_arg_mainFileNameWithExtension])) {
    output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_arg_mainFileNameWithExtension . "' missing");
    exit(1);
}
# the main file name
$arg_mainFileNameWithExtension = $body[$s_arg_mainFileNameWithExtension];


if (!isset($body[$s_arg_allFiles])) {
    output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_arg_allFiles . "' missing");
    exit(1);
}
# all files to use
$arg_allFiles = $body[$s_arg_allFiles];


# check/ensure the iteration dir for the current iteration

$hardGlobalTimeoutInMs = $config[$s_arg_hardTimeoutInMs];

if ($hardGlobalTimeoutInMs === NULL) {
    output($status_code_InternalServerError, 'could not read global hard timeout in ms from config');
    exit(1);
}

$hardGlobalCompileTimeoutInMs = $config[$s_arg_hardCompileTimeoutInMs];

if ($hardGlobalCompileTimeoutInMs === NULL) {
    output($status_code_InternalServerError, 'could not read global hard compile timeout in ms from config');
    exit(1);
}

$hardMemoryLimitInKb = $config[$s_arg_hardMemoryLimitInKb];

if ($hardMemoryLimitInKb === NULL) {
    output($status_code_InternalServerError, 'could not read global hard memory limit in kb from config');
    exit(1);
}

$hardDiskSpaceLimitInKb = $config[$s_arg_hardDiskSpaceLimitInKb];

if ($hardDiskSpaceLimitInKb === NULL) {
    output($status_code_InternalServerError, 'could not read global hard disk space limit in kb from config');
    exit(1);
}

if (!isset($config[$s_arg_insertTraceLogs])) {
    output($status_code_InternalServerError, 'could not find insert trace log setting in config');
    exit(1);
}

$insertTraceLogs = $config[$s_arg_insertTraceLogs];

if ($insertTraceLogs === NULL) {
    output($status_code_InternalServerError, 'could not read insert trace log from config');
    exit(1);
}

$dbTraceTableName = $config[$s_arg_dbTraceTableName];

if ($dbTraceTableName === NULL) {
    output($status_code_InternalServerError, 'could not read the trace table from config');
    exit(1);
}


if (!isset($body[$s_requestTraceString])) {
    output($status_code_InternalServerError, 'request trace string was not found in the request');
    exit(1);
}
$traceString = $body[$s_requestTraceString];


$requestDistinctionString = NULL;
# try to get the request distinction key/number
if (!isset($body[$s_requestDistinctionNumber])) {
    # client did not set a request distinction number...
    # ignore and hope microtime will handle...
} else {

    if (is_valid_fileName($body[$s_requestDistinctionNumber])) {
        # a valid file name as request id (used to create a dir)
        $requestDistinctionString = $body[$s_requestDistinctionNumber];
    } else {
        output($status_code_InternalServerError, 'request distinct number is not a regex with only letters, numbers and underscores');
        exit(1);
    }
}


if ($isDebug) {
    $time_post = microtime(true);
    $exec_time = ($time_post - $time_pre) * 1000; //in ms
    debug("time to decode body: " . $exec_time);
}

if ($isDebug) {
    $time_pre = microtime(true);
}
$rootDirName = $rootDirNameToUseString;

if ($rootDirName === NULL) {
    goto handleCleanupWithoutDir;
}

if ($isDebug) {
    $time_post = microtime(true);
    $exec_time = ($time_post - $time_pre) * 1000; //in ms
    debug("time to check the next root dir: " . $exec_time);
}

if ($isDebug) {
    $time_pre = microtime(true);
}
# get the transaction
$conn = connectToDb($config['dbServer'], $config['dbName'], $config['dbUser'], $config['dbPw']);

if ($conn === NULL) {
    output($status_code_DbIssue, 'could not connect to db');
    exit(1);
}

$transactionKey = startTransaction($conn, $config, $rootDirName, 'var', $requestDistinctionString);

if ($isDebug) {
    $time_post = microtime(true);
    $exec_time = ($time_post - $time_pre) * 1000; //in ms
    debug("time to open db connection & start the transaction: " . $exec_time);
}

# other arguments depends on the command argument


if (!isset($body[$s_arg_pLang])) {
    endTransaction($conn, $config, $transactionKey, $rootDirName);
    output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_arg_pLang . "' missing");
    exit(1);
}

if (!isset($body[$s_arg_pLang])) {
    endTransaction($conn, $config, $transactionKey, $rootDirName);
    output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_arg_pLang . "' missing");
    exit(1);
}

if (!isset($body[$s_arg_characterLimitProtocol])) {
    endTransaction($conn, $config, $transactionKey, $rootDirName);
    output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_arg_characterLimitProtocol . "' missing");
    exit(1);
}

$arg_pLang = $body[$s_arg_pLang];

$arg_characterLimitProtocol = $body[$s_arg_characterLimitProtocol];


# check if the p lang is supported...
# ...get the script to execute from the db like domjudge

if (isset($config['dbPLangTableName']) === FALSE) {
    $config['dbPLangTableName'] = $config['dbTableName']; //backwards compatibility
}

$sql = "SELECT * FROM `" . $config['dbPLangTableName'] . "`;";
$allRows = executeSqlAndReturnValue($conn, $sql, []);
$compileCmd = NULL;
$execCmd = NULL;
$sourceFileExtensions = [];

foreach ($allRows as $row) {

    if ($row[$s_plang_internalName] === $arg_pLang) {
        //ok found p lang commands
        $compileCmd = $row[$s_compileCommand];
        $execCmd = $row[$s_executeCommand];

        $sourceFileExtensions = json_decode($row[$s_source_file_extensions], TRUE);

        if ($sourceFileExtensions === NULL) {
            if ($isDebug) {
                fwrite(STDOUT, 'could not decode extensions row');
            }
            endTransaction($conn, $config, $transactionKey, $rootDirName);
            output($status_code_DbIssue, "invalid extensions");
            exit(1);
        }
        break;
    }
}


# create source files
$rootDir = $rootDirName;
$paths = ensureDirs($rootDir, $transactionKey);

if ($paths === FALSE) {
    endTransaction($conn, $config, $transactionKey, $rootDirName);
    output($status_code_FileSystemIssue, "could not create dirs");
    exit(1);
}

$fullWorkingDirPath = $paths[$s_fullWorkingDirPath];

if (createFiles($arg_allFiles, $fullWorkingDirPath) === FALSE) {
    endTransaction($conn, $config, $transactionKey, $rootDirName);
    //error message is in createFiles function
    exit(1);
}


# create a user program source code variable (for logging)

$allUserSourceCode = '';

foreach ($arg_allFiles as $file) {

    $_fileName  = $file[$s_fileName];
    $_fileContent = $file[$s_fileContent];

    $allUserSourceCode = $allUserSourceCode . "---" . $_fileName . "---\n"  . $_fileContent . "\n\n";
}



if ($isDebug) {
    debug("all files created!");
}

$testCases = $body['tests'];
$testResults = [];

$testCount = 0;

//check if we already tied to compile
//if the first test is a regex test it wont compile...
$hasTriedToCompile = FALSE;

//store the compile result in case we only copy the result (output for other tests)
$compileResult = array(); //the result of the test
$compileRawResult = array(); //the raw result from do blackbox test

foreach ($testCases as $test) {

    if ($testCount >= $config[$s_arg_maxNumberOfTestsWithOneRequest]) {
        break;
    }

    $testId = $test[$s_test_id];
    $testContent = $test[$s_test_content]; //can be null or empty
    $testAssets = $test[$s_test_allAssets];
    $compilerOptions = $test[$s_test_CompilerOptions]; //can be null or empty


    if (!isset($test[$s_test_command]) || !isset($test[$s_test_id])
        //its ok if we have test content
        //but throw if content is null (empty string is ok)
        || (!isset($test[$s_test_content]) && $test[$s_test_command] !== $s_command_compile)
    ) {

        #output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_test_command . "' missing");
        $missingArgument = $s_test_command;

        if (!isset($test[$s_test_command])) {
            $missingArgument = $s_test_command;
        } else if (!isset($test[$s_test_content])) {
            $missingArgument = $s_test_content;
        } else if (isset($test[$s_test_id])) {
            $missingArgument = $s_test_id;
        }

        $result = create_test_result($status_code_ArgumentInvalidOrMissing, "argument '" . $missingArgument . "' missing", $testId,
            FALSE, FALSE, NULL,
            NULL, NULL, NULL,
            100, '',
            NULL, NULL,
            false, $arg_characterLimitProtocol
        );

        array_push($testResults, $result);
        continue;
    }

    # the command to execute
    $arg_command = $test[$s_test_command];

    //check if valid/supported command
    if (!in_array($arg_command, $s_supportedCommands)) {

        $result = create_test_result($status_code_ArgumentInvalidOrMissing, "unknown command, supported commands: [" . implode(', ', $s_supportedCommands) . ']', $testId,
            FALSE, FALSE, NULL,
            NULL, NULL, NULL,
            100, '',
            NULL, NULL,
            false, $arg_characterLimitProtocol
        );
        array_push($testResults, $result);
        continue;
    }

    if ($arg_command === $s_command_compile) {


        if ($compileCmd === NULL) {
            $result = create_test_result($status_code_DbIssue, 'could not find compile command', $testId,
                FALSE, FALSE, NULL,
                NULL, NULL, NULL,
                100, '',
                NULL, NULL,
                false, $arg_characterLimitProtocol
            );
            array_push($testResults, $result);
            continue;
        } else {

            require_once './do_compileTest_func.php';

            $showTestRunnerDebugOutput = FALSE;
            if (isset($config['showTestRunnerDebugOutput'])) {
                $showTestRunnerDebugOutput = $config['showTestRunnerDebugOutput'];
            }

            if (!isset($test[$s_test_compileTimeoutInMs])) {
                $result = create_test_result($status_code_ArgumentInvalidOrMissing, "argument '" . $s_test_compileTimeoutInMs . "' missing", $testId,
                    FALSE, FALSE, NULL,
                    NULL, NULL, NULL,
                    100, '',
                    NULL, NULL,
                    false, $arg_characterLimitProtocol
                );
                array_push($testResults, $result);
                continue;
            }

            $arg_compileTimeout = $test[$s_test_compileTimeoutInMs];

            $min_compileTimeout = $hardGlobalCompileTimeoutInMs;

            if ($arg_compileTimeout) {
                $min_compileTimeout = min($arg_compileTimeout, $hardGlobalCompileTimeoutInMs);
            }

            try {
                $_result = do_compile($conn, $traceString, $allUserSourceCode, $arg_mainFileNameWithExtension, $fullWorkingDirPath, $min_compileTimeout,
                    $compileCmd, $sourceFileExtensions, $requestDistinctionString, $showTestRunnerDebugOutput, $compilerOptions,
                    $arg_characterLimitProtocol);
            } catch (Exception $e) {
                //critical error, abort execution (all other tests are likely to fail)
                output($status_code_InternalServerError, "error executing regex test: " . $e->getMessage());
                goto handleCleanup;
            }

            $result = create_test_result($status_code_Ok, '', $testId,
                $_result[$s_return_val] === 0, $_result[$s_return_val] !== 50 && $_result[$s_return_val] !== 7, $_result[$s_user_program_exit_code],
                $_result[$s_timeForCompiling], $_result[$s_timeForUserProgram], $_result[$s_runnerVersion],
                $_result[$s_return_val], formatOutput($_result[$s_output]),
                $_result[$s_compileTimeoutInMsUsed], NULL,
                $_result[$s_characterLimitExceeded], $arg_characterLimitProtocol
            );
            array_push($testResults, $result);

            //even if the compilation failed we tried so set $hasCompiled to true (others would faild too)
            $hasTriedToCompile = TRUE;

        }
    } else if ($arg_command === $s_command_regexTest) {

        require_once './do_regexTest_func.php';

        $showTestRunnerDebugOutput = FALSE;
        if (isset($config['showTestRunnerDebugOutput'])) {
            $showTestRunnerDebugOutput = $config['showTestRunnerDebugOutput'];
        }

        try {
            //no need to write any files ... can all be handled in memory
            //$_result = do_regexTest_json($arg_allFiles, $testContent);
            $_result = do_regexTest_format2($arg_allFiles, $testContent, $showTestRunnerDebugOutput, $arg_characterLimitProtocol);
        } catch (Exception $e) {
            //critical error, abort execution (all other tests are likely to fail)
            output($status_code_InternalServerError, "error executing regex test: " . $e->getMessage());
            goto handleCleanup;
        }

        $result = create_test_result($status_code_Ok, '', $testId,
            $_result[$s_return_val] === 0, $_result[$s_return_val] !== 50 && $_result[$s_return_val] !== 7, NULL,
            NULL, NULL, NULL,
            $_result[$s_return_val], $_result[$s_output],
            NULL, NULL,
            $_result[$s_characterLimitExceeded], $arg_characterLimitProtocol
        );
        array_push($testResults, $result);

    } else if ($arg_command === $s_command_blackBoxTest || $arg_command === $s_command_justRunTest || $arg_command === $s_command_compareFilesTest) {

        if ($isDebug) {
            $time_pre = microtime(true);
        }

        if (!isset($test[$s_test_timeoutInMs])) {

            $result = create_test_result($status_code_ArgumentInvalidOrMissing, "argument '" . $s_test_timeoutInMs . "' missing", $testId,
                FALSE, FALSE, NULL,
                NULL, NULL, NULL,
                100, '',
                NULL, NULL,
                false, $arg_characterLimitProtocol

            );
            array_push($testResults, $result);
            continue;
        }

        if (!isset($test[$s_test_compileTimeoutInMs])) {

            $result = create_test_result($status_code_ArgumentInvalidOrMissing, "argument '" . $s_test_compileTimeoutInMs . "' missing", $testId,
                FALSE, FALSE, NULL,
                NULL, NULL, NULL,
                100, '',
                NULL, NULL,
                false, $arg_characterLimitProtocol
            );
            array_push($testResults, $result);
            continue;
        }

        # the timeout for the test
        $arg_timeout = $test[$s_test_timeoutInMs];
        $arg_compileTimeout = $test[$s_test_compileTimeoutInMs];


        if (!isset($test[$s_test_memoryLimitInKb])) {

            $result = create_test_result($status_code_ArgumentInvalidOrMissing, "argument '" . $s_test_memoryLimitInKb . "' missing", $testId,
                FALSE, FALSE, NULL,
                NULL, NULL, NULL,
                100, '',
                NULL, NULL,
                false, $arg_characterLimitProtocol
            );
            array_push($testResults, $result);
            continue;
        }

        # the memory limit for the test
        $arg_MemoryLimit = $test[$s_test_memoryLimitInKb];


        if (!isset($test[$s_test_diskSpaceLimitInKb])) {

            $result = create_test_result($status_code_ArgumentInvalidOrMissing, "argument '" . $s_test_diskSpaceLimitInKb . "' missing", $testId,
                FALSE, FALSE, NULL,
                NULL, NULL, NULL,
                100, '', $versionString,
                false, $arg_characterLimitProtocol
            );
            array_push($testResults, $result);
            continue;
        }

        # the max dis space the test can write to
        $arg_DiskSpaceLimit = $test[$s_test_diskSpaceLimitInKb];


        if ($compileCmd === NULL || $execCmd === NULL) {
            $result = create_test_result($status_code_DbIssue, "could not find compile command or execute command (unknown plang?)", $testId,
                FALSE, FALSE, NULL,
                NULL, NULL, NULL,
                100, '',
                NULL, NULL,
                false, $arg_characterLimitProtocol
            );
            array_push($testResults, $result);
            continue;
        }

        # ensure the hard limits

        $min_timeout = $hardGlobalTimeoutInMs;

        if ($arg_timeout) {
            $min_timeout = min($arg_timeout, $hardGlobalTimeoutInMs);
        }

        $min_compileTimeout = $hardGlobalCompileTimeoutInMs;

        if ($arg_compileTimeout) {
            $min_compileTimeout = min($arg_compileTimeout, $hardGlobalCompileTimeoutInMs);
        }

        $min_memoryLimit = $hardMemoryLimitInKb;

        if ($arg_MemoryLimit) {
            $min_memoryLimit = min($arg_MemoryLimit, $hardMemoryLimitInKb);
        }

        $min_diskSpaceLimit = $hardDiskSpaceLimitInKb;

        if ($arg_DiskSpaceLimit) {
            $min_diskSpaceLimit = min($arg_DiskSpaceLimit, $hardDiskSpaceLimitInKb);
        }


        # was just a test...
        #output("t" . $min_timeout . " m" . $min_memoryLimit . " d" . $min_diskSpaceLimit . " dd: " . $arg_DiskSpaceLimit);

        if ($isDebug) {
            $time_post = microtime(true);
            $exec_time = ($time_post - $time_pre) * 1000; //in ms
            debug("time get all args for the black-box-test: " . $exec_time);
        }

        //only the first test needs to compile sources ...

        $needCompilation = $hasTriedToCompile ? FALSE : TRUE;

        $needCompilation = !$hasTriedToCompile;


        if ($needCompilation === FALSE && $compileResult[$s_hasCompiled] === FALSE) {
            //if the compile fail ... all other tests would give use main file not found...
            //so just copy the compile error from the failed compiled test
            $_result = $compileRawResult;
        } else {


            $showTestRunnerDebugOutput = FALSE;
            if (isset($config['showTestRunnerDebugOutput'])) {
                $showTestRunnerDebugOutput = $config['showTestRunnerDebugOutput'];
            }

            if ($arg_command === $s_command_compareFilesTest) {

                require_once './do_compareFilesTest_func.php';

                try {
                    $_result = do_compareFilesTest($conn, $traceString, $allUserSourceCode, $arg_mainFileNameWithExtension, $test, $fullWorkingDirPath,
                        $min_timeout, $min_compileTimeout, $min_memoryLimit, $min_diskSpaceLimit,
                        $compileCmd, $execCmd, $sourceFileExtensions, $needCompilation,
                        $requestDistinctionString,
                        $showTestRunnerDebugOutput,
                        $compilerOptions,
                        $arg_characterLimitProtocol
                    );
                } catch (Exception $e) {
                    //critical error, abort execution (all other tests are likely to fail)
                    output($status_code_InternalServerError, "error executing compare files test: " . $e->getMessage());
                    goto handleCleanup;
                }
            } else {
                require_once './do_blackBoxTest_func.php';

                try {
                    $_result = do_blackBoxTest($conn, $traceString, $allUserSourceCode, $arg_mainFileNameWithExtension, $test, $fullWorkingDirPath,
                        $min_timeout, $min_compileTimeout, $min_memoryLimit, $min_diskSpaceLimit,
                        $compileCmd, $execCmd, $sourceFileExtensions, $needCompilation,
                        ($arg_command === $s_command_justRunTest),
                        $requestDistinctionString,
                        $showTestRunnerDebugOutput,
                        $compilerOptions,
                        $arg_characterLimitProtocol
                    );
                } catch (Exception $e) {
                    //critical error, abort execution (all other tests are likely to fail)
                    output($status_code_InternalServerError, "error executing black box test: " . $e->getMessage());
                    goto handleCleanup;
                }
            }

        }


        $result = create_test_result($status_code_Ok, '', $testId,
            $_result[$s_return_val] === 0, $_result[$s_return_val] !== 50 && $_result[$s_return_val] !== 7, $_result[$s_user_program_exit_code],
            $_result[$s_timeForCompiling], $_result[$s_timeForUserProgram], $_result[$s_runnerVersion],
            $_result[$s_return_val], formatOutput($_result[$s_output]),
            $_result[$s_compileTimeoutInMsUsed], $_result[$s_timeoutInMsUsed],
            $_result[$s_characterLimitExceeded], $arg_characterLimitProtocol
        );

        if ($needCompilation) {
            //this tried to compile ... store result
            $compileResult = $result;
            $compileRawResult = $_result;
        }

        array_push($testResults, $result);

        //even if the compilation failed we tried so set $hasCompiled to true (others would fail too)
        $hasTriedToCompile = TRUE;
    }

    $testCount++;
}

handleCleanup:
deleteDir($fullWorkingDirPath);
handleCleanupWithoutDir:
# release the transaction
endTransaction($conn, $config, $transactionKey, $rootDirName);

output($status_code_Ok, '', $testResults);

